@@ -29,7 +29,6 @@ static void gail_widget_connect_widget_destroyed (GtkAccessible    *accessible);
 static void gail_widget_destroyed                (GtkWidget        *widget,
                                                   GtkAccessible    *accessible);
 
-static G_CONST_RETURN gchar* gail_widget_get_name (AtkObject *accessible);
 static G_CONST_RETURN gchar* gail_widget_get_description (AtkObject *accessible);
 static AtkObject* gail_widget_get_parent (AtkObject *accessible);
 static AtkStateSet* gail_widget_ref_state_set (AtkObject *accessible);
@@ -156,7 +155,6 @@ gail_widget_class_init (GailWidgetClass *klass)
 
   accessible_class->connect_widget_destroyed = gail_widget_connect_widget_destroyed;
 
-  class->get_name = gail_widget_get_name;
   class->get_description = gail_widget_get_description;
   class->get_parent = gail_widget_get_parent;
   class->ref_relation_set = gail_widget_ref_relation_set;
@@ -260,30 +258,6 @@ gail_widget_destroyed (GtkWidget     *widget,
                                   TRUE);
 }
 
-static G_CONST_RETURN gchar*
-gail_widget_get_name (AtkObject *accessible)
-{
-  if (accessible->name)
-    return accessible->name;
-  else
-    {
-      /*
-       * Get the widget name is it exists
-       */
-      GtkWidget* widget = GTK_ACCESSIBLE (accessible)->widget;
-
-      if (widget == NULL)
-        /*
-         * State is defunct
-         */
-        return NULL;
-
-      g_return_val_if_fail (GTK_IS_WIDGET (widget), NULL);
-
-      return widget->name;
-    }
-}
-
 static G_CONST_RETURN gchar*
 gail_widget_get_description (AtkObject *accessible)
 {
@@ -345,7 +319,8 @@ gail_widget_get_parent (AtkObject *accessible)
 
 static GtkWidget*
 find_label (GtkWidget *start,
-            GtkWidget *widget)
+            GtkWidget *widget,
+            gboolean  search_all)
 {
 
   GList *list, *l;
@@ -378,17 +353,87 @@ find_label (GtkWidget *start,
                 return NULL;
             }
           g_return_val_if_fail (l, NULL);
-          if (l->prev && GTK_IS_LABEL (l->prev->data))
-	    {  
-              label = GTK_WIDGET (l->prev->data);
-              if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                label = NULL;
-            } 
-          if (!label && l->next && GTK_IS_LABEL (l->next->data))
-	    {  
-              label = GTK_WIDGET (l->next->data);
-              if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
-                label = NULL;
+          if (search_all)
+            {
+              GList *temp_l;
+
+              temp_l = l;
+              while (temp_l->prev)
+                {
+                  if (GTK_IS_LABEL (temp_l->prev->data))
+	            {  
+                      label = GTK_WIDGET (temp_l->prev->data);
+                      if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
+                        label = NULL;
+                      else
+                        break;
+                    }
+                  /* Heuristic to find label in GtkFileSelection */
+                  else if (GTK_IS_EVENT_BOX (temp_l->prev->data))
+                    {
+                       label = GTK_BIN (temp_l->prev->data)->child;
+                       if (GTK_IS_LABEL (label) &&
+                           gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
+                         label = NULL;
+                       else
+                         break;
+                    }
+                  temp_l = temp_l->prev;
+                }
+              if (label == NULL)
+                {
+                  temp_l = l;
+                  while (temp_l->next)
+                    {
+                      if (GTK_IS_LABEL (temp_l->next->data))
+	                {  
+                          label = GTK_WIDGET (temp_l->next->data);
+                          if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
+                            label = NULL;
+                          else
+                            break;
+                        }
+                      /* Heuristic to find label in NautilusFilePropertiesWindow
+                         bug #134629 */
+                      if (GTK_IS_HBOX (temp_l->next->data))
+                        {
+                          GList *box_children, *temp_children;
+
+                          temp_children = box_children = gtk_container_get_children (GTK_CONTAINER (temp_l->next->data));
+                          while (temp_children)
+                            {
+                              if (GTK_IS_LABEL (temp_children->data))
+                                {
+                                  label = GTK_WIDGET (temp_children->data);
+                                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
+                                    label = NULL;
+                                  else
+                                    break;
+                                }
+                              temp_children = temp_children->next;
+                            }
+                          g_list_free (box_children);
+                          if (label)
+                            break;
+                        }
+                      temp_l = temp_l->next;
+                    }
+                }
+            }
+          else 
+            { 
+              if (l->prev && GTK_IS_LABEL (l->prev->data))
+	        {  
+                  label = GTK_WIDGET (l->prev->data);
+                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
+                    label = NULL;
+                }
+              if (!label && l->next && GTK_IS_LABEL (l->next->data))
+	        {  
+                  label = GTK_WIDGET (l->next->data);
+                  if (gtk_label_get_mnemonic_widget (GTK_LABEL (label)) != widget)
+                    label = NULL;
+                }
             } 
 
           g_list_free (list);
@@ -417,26 +462,104 @@ gail_widget_ref_relation_set (AtkObject *obj)
 
   relation_set = ATK_OBJECT_CLASS (parent_class)->ref_relation_set (obj);
 
+  if (GTK_IS_BOX (widget) && !GTK_IS_COMBO (widget))
+      /*
+       * Do not report labelled-by for a GtkBox which could be a 
+       * GnomeFileEntry.
+       */
+    return relation_set;
+
   if (!atk_relation_set_contains (relation_set, ATK_RELATION_LABELLED_BY))
     {
       /*
        * Heuristic to create LABELLED_BY relation:
        *
-       * If the previous or next child of the current widget is a label 
+       * If any sibling the current widget is a label 
        * and its mnemonic widget is the current widget then set LABELLED_BY 
        * relation on current accessible object.
        *
        * If this fails try to find a label at the same level as the widget's
        * parent. This finds labels for NautilusProgramChooser dialog.
        */
-      label = find_label (widget, widget);
-      if (!label) 
+      label = find_label (widget, widget, TRUE);
+      if (label == NULL)
         {
-          GtkWidget *parent;
+          GtkWidget *temp_widget;
 
-          parent = gtk_widget_get_parent (widget);
-          if (parent)
-            label = find_label (parent, widget);
+          temp_widget = gtk_widget_get_parent (widget);
+          if (temp_widget)
+            label = find_label (temp_widget, widget, FALSE);
+          /*
+           * This finds relationship in Desktop Preferences Background.
+           * See bug #126973.
+           * This finds relationship for GnomeEntry inside a GnomeFileEntry
+	   * and its label.
+           * See bug #133403.
+           */
+          if (!label)
+            {
+              if ((GTK_IS_VBOX (temp_widget) && GTK_IS_OPTION_MENU (widget)) ||
+                  (GTK_IS_HBOX (temp_widget) && GTK_IS_COMBO (widget)))
+                {
+                  temp_widget = gtk_widget_get_parent (temp_widget);
+                  if (temp_widget)
+                    {
+                      if (GTK_IS_OPTION_MENU (widget))
+                        label = find_label (temp_widget, widget, FALSE);
+                      else if (GTK_IS_COMBO (widget))
+                        label = find_label (temp_widget, temp_widget, FALSE);
+                      if (!label)
+                        {
+                          /*
+                           * Find label for combo-box in Sound preferences
+                           * capplet. See bug #126972.
+                           */
+                          temp_widget = gtk_widget_get_parent (temp_widget);
+                          if (GTK_IS_VBOX (temp_widget))
+                            {
+                              GList *list;
+
+                              list = gtk_container_get_children (GTK_CONTAINER (temp_widget));
+                              if (list)
+                                {
+                                  temp_widget = list->data;
+                                  if (GTK_IS_VBOX (temp_widget))
+                                    {
+                                      g_list_free (list);
+                                      list = gtk_container_get_children (GTK_CONTAINER (temp_widget));
+                                      if (list)
+                                        {
+                                          temp_widget = list->data;
+                                          g_list_free (list);
+                                          if (GTK_IS_LABEL (temp_widget) &&
+                                              gtk_label_get_mnemonic_widget (GTK_LABEL (temp_widget)) == widget)
+                                            {
+                                              label = temp_widget; 
+                                            }
+                                        }
+                                    }
+                                }
+                            }
+
+                        }
+                    }
+                }
+              /*
+               * Handle the case where GnomneIconEntry is the mnemonic widget.
+               * The GtkButton which is a grandchild of the GnomeIconEntry
+               * should really be the mnemonic widget. See bug #133967.
+               */
+              else if (GTK_IS_BUTTON (widget) && GTK_IS_ALIGNMENT (temp_widget))
+                {
+                  temp_widget = gtk_widget_get_parent (temp_widget);
+                  if (GTK_IS_BOX (temp_widget))
+                    {
+                      label = find_label (temp_widget, temp_widget, TRUE);
+                      if (!label)
+                        label = find_label (gtk_widget_get_parent (temp_widget), temp_widget, FALSE);
+                    }
+                }
+            }
         }
       if (label)
         {
@@ -1039,17 +1162,7 @@ static gboolean gail_widget_on_screen (GtkWidget *widget)
           widget->allocation.y + widget->allocation.height <= 0)
         return_value = FALSE;
       else 
-        {
-        /*
-         * Check whether the widget has been placed of the screen. The
-         * widget may be MAPPED as when toolbar items do not fit on the toolbar.
-         */
-          if (widget->allocation.x + widget->allocation.width <= 0 &&
-              widget->allocation.y + widget->allocation.height <= 0)
-            return_value = FALSE;
-          else 
-            return_value = TRUE;
-        }
+        return_value = TRUE;
     }
 
   return return_value;
